/*
 *	PIC 12C5xx DISASSEMBLER FOR MODCHIP Ver0.11
 *	COPYRIGHT(C) 1999 Hanimar
 *	This program can be freely distributed.
 */
#include <stdio.h>
#include <string.h>
#define IAMTHEOWNER
#include "pic12c5x.h"
#undef IAMTHEOWNER

static char LastErrorString[256] = "No error.";

char *Mnemonics[] = {
	"UNKNOWN",
	"addwf" , "andwf" , "clrf" , "clrw" ,
	"comf", "decf" , "decfsz" , "incf" ,
	"incfsz" , "iorwf" , "movf" , "movwf" ,
	"nop" , "rlf" , "rrf" , "subwf" ,
	"swapf" , "xorwf" , "bcf" , "bsf" ,
	"btfsc" , "btfss" , "andlw" , "call" ,
	"clrwdt" , "goto" , "iorlw" , "movlw" ,
	"option" , "retlw" , "sleep" , "tris" ,
	"xorlw" ,
	NULL
};

/*
 *	Returns descriptions of last error.
 */
const char *PicGetLastError(void) { return LastErrorString; }

/*
 *	Flash ROM image.
 */
void PicFlash(PIC12C5XX *pMod, const char *fname, ROMTYPE romType)
{
	memset(pMod , 0 , sizeof (PIC12C5XX));
	pMod->romType = romType;

	strncpy(pMod->fname, fname, sizeof(pMod->fname));
	pMod->fname[sizeof(pMod->fname) -1] = '\0';

	strcpy(LastErrorString, "No error.");
	return;
}

/*
 *	Write opcode into address.
 *	Returns:
 *		 0: success
 *		-1:	error. Call PicGetLastError()
 */
int PicWrite(PIC12C5XX *pMod, int address, int opcode)
{
	int ret = 0;
	int upper;
	int lower;
	OPCODE *pOp;
	int i;

	if (address < 0 || address >= C509_ROMSIZE) {
		sprintf(LastErrorString , "ERROR:Invalid address %04X specified.",
			address);
		return -1;
	}
	if (pMod->romType == C508 && address >= C508_ROMSIZE) {
		sprintf(LastErrorString ,
			"WARNING: Invalid address for 12C508 specified",
			address);
		ret = 1;
	}

	pOp = &pMod->rom[address];
	pOp->code = opcode;
	pOp->flags |= FLAG_USED;

	upper = (opcode & 0xff00) >> 8;
	lower = opcode & 0xff;
	switch (upper) {
	case 0x00:
		/*
		 *	MOVWF	f	0000 001f ffff
		 *	CLRF	f	0000 011f ffff
		 *	DECF	f,d	0000 11df ffff
		 *	SUBWF	f,d	0000 10df ffff
		 *	NOP		-	0000 0000 0000
		 *	CLRW	-	0000 0100 0000
		 *	OPTION	-	0000 0000 0010
		 *	SLEEP	-	0000 0000 0011
		 *	CLRWDT	-	0000 0000 0100
		 *	TRIS	f	0000 0000 0fff
		 */
		if ((lower >> 5) == 1) {
			pOp->mnemonic = MN_MOVWF;
			pOp->op1 = lower & 0x1f; pOp->op2 = OPERAND_UNUSED;
			pMod->ram[pOp->op1] = RAM_USED;
		} else if ((lower >> 5) == 3) {
			pOp->mnemonic = MN_CLRF;
			pOp->op1 = lower & 0x1f; pOp->op2 = OPERAND_UNUSED;
			pMod->ram[pOp->op1] = RAM_USED;
		} else if ((lower >> 6) == 2) {
			pOp->mnemonic = MN_SUBWF;
			pOp->op1 = lower & 0x1f; pOp->op2 = (lower >> 5) & 1;
			pMod->ram[pOp->op1] = RAM_USED;
		} else if ((lower >> 6) == 3) {
			pOp->mnemonic = MN_DECF;
			pOp->op1 = lower & 0x1f; pOp->op2 = (lower >> 5) & 1;
			pMod->ram[pOp->op1] = RAM_USED;
		} else if (lower == 0) {
			pOp->mnemonic = MN_NOP;
			pOp->op1 = pOp->op2 = OPERAND_UNUSED;
		} else if (lower == 0x40) {
			pOp->mnemonic = MN_CLRW;
			pOp->op1 = pOp->op2 = OPERAND_UNUSED;
		} else if (lower == 2) {
			pOp->mnemonic = MN_OPTION;
			pOp->op1 = pOp->op2 = OPERAND_UNUSED;
		} else if (lower == 3) {
			pOp->mnemonic = MN_SLEEP;
			pOp->op1 = pOp->op2 = OPERAND_UNUSED;
		} else if (lower == 4) {
			pOp->mnemonic = MN_CLRWDT;
			pOp->op1 = pOp->op2 = OPERAND_UNUSED;
		} else if ((lower >> 3) == 0) {	/* This must be last */
			pOp->mnemonic = MN_TRIS;
			pOp->op1 = lower & 0x07; pOp->op2 = OPERAND_UNUSED;
		} else {
			sprintf(LastErrorString ,
				"ERROR: Invalid opcode %0x in address %04X.",
				opcode,
				address
				);
			pOp->mnemonic = MN_UNKNOWN;
			pOp->op1 = pOp->op2 = OPERAND_UNUSED;
		}
		break;
	
	case 0x01:
		/*
		 *	ADDWF	f,d	0001 11df ffff
		 *	ANDWF	f,d	0001 01df ffff
		 *	IORWF	f,d	0001 00df ffff
		 *	XORWF	f,d	0001 10df ffff
		 */
		i = lower >> 6;
		if (i == 3) pOp->mnemonic = MN_ADDWF;
		else if (i == 1) pOp->mnemonic = MN_ANDWF;
		else if (i == 0) pOp->mnemonic = MN_IORWF;
		else pOp->mnemonic = MN_XORWF;

		pOp->op1 = lower & 0x1f; pOp->op2 = (lower >> 5) & 1;
		pMod->ram[pOp->op1] = RAM_USED;
		break;
	
	case 0x02:
		/*
		 *	DECFSZ	f,d	0010 11df ffff
		 *	INCF	f,d	0010 10df ffff
		 *	MOVF	f,d	0010 00df ffff
		 *	COMF	f,d	0010 01df ffff
		 */
		i = lower >> 6;
		if (i == 3) {
			pOp->mnemonic = MN_DECFSZ;
			pOp->flags |= FLAG_BRANCH;
		} else if (i == 2) pOp->mnemonic = MN_INCF;
		else if (i == 0) pOp->mnemonic = MN_MOVF;
		else pOp->mnemonic = MN_COMF;

		pOp->op1 = lower & 0x1f; pOp->op2 = (lower >> 5) & 1;
		pMod->ram[pOp->op1] = RAM_USED;
		break;
	
	case 0x03:
		/*
		 *	INCFSZ	f,d	0011 11df ffff
		 *	RRF		f,d	0011 00df ffff
		 *	RLF		f,d	0011 01df ffff
		 *	SWAPF	f,d	0011 10df ffff
		 */
		i = lower >> 6;
		if (i == 3) {
			pOp->mnemonic = MN_INCFSZ;
			pOp->flags |= FLAG_BRANCH;
		} else if (i == 0) pOp->mnemonic = MN_RRF;
		else if (i == 1) pOp->mnemonic = MN_RLF;
		else pOp->mnemonic = MN_SWAPF;

		pOp->op1 = lower & 0x1f; pOp->op2 = (lower >> 5) & 1;
		pMod->ram[pOp->op1] = RAM_USED;
		break;
	
	case 0x04:
		/*
		 *	BCF		f,d	0100 bbbf ffff
		 */
		pOp->mnemonic = MN_BCF;
		pOp->op1 = lower & 0x1f;
		pOp->op2 = (lower >> 5) & 0x07;

		if (pOp->op1 == REG_GPIO) pMod->gpioFlag |= (1 << pOp->op2);
		pMod->ram[pOp->op1] = RAM_USED;
		break;
	
	case 0x05:
		/*
		 *	BSF		f,d	0101 bbbf ffff
		 */
		pOp->mnemonic = MN_BSF;
		pOp->op1 = lower & 0x1f;
		pOp->op2 = (lower >> 5) & 0x07;

		if (pOp->op1 == REG_GPIO) pMod->gpioFlag |= (1 << pOp->op2);
		pMod->ram[pOp->op1] = RAM_USED;
		break;
	
	case 0x06:
		/*
		 *	BTFSC	f,d	0110 bbbf ffff
		 */
		pOp->mnemonic = MN_BTFSC;
		pOp->op1 = lower & 0x1f;
		pOp->op2 = (lower >> 5) & 0x07;
		pOp->flags |= FLAG_BRANCH;

		if (pOp->op1 == REG_GPIO) pMod->gpioFlag |= (1 << pOp->op2);
		pMod->ram[pOp->op1] = RAM_USED;

		break;
	
	case 0x07:
		/*
		 *	BTFSS	f,d	0111 bbbf ffff
		 */
		pOp->mnemonic = MN_BTFSS;
		pOp->op1 = lower & 0x1f;
		pOp->op2 = (lower >> 5) & 0x07;
		pOp->flags |= FLAG_BRANCH;

		if (pOp->op1 == REG_GPIO) pMod->gpioFlag |= (1 << pOp->op2);
		pMod->ram[pOp->op1] = RAM_USED;
		break;
	
	case 0x08:
		/*
		 *	RETLW	k	1000 kkkk kkkk
		 */
		pOp->mnemonic = MN_RETLW;
		pOp->op1 = lower;
		pOp->op2 = OPERAND_UNUSED;
		break;
	
	case 0x09:
		/*
		 *	CALL	k	1001 kkkk kkkk
		 */
		pOp->mnemonic = MN_CALL;
		pOp->op1 = lower;
		pOp->op2 = OPERAND_UNUSED;
		
		pMod->rom[pOp->op1].flags |= FLAG_CALLED;
		break;

	case 0x0a:	/* PASS THRU */
	case 0x0b:
		/*
		 *	GOTO	k	101k kkkk kkkk
		 */
		pOp->mnemonic = MN_GOTO;
		pOp->op1 = opcode - 0x0a00;
		pOp->op2 = OPERAND_UNUSED;

		if (pOp->op1 >= C509_ROMSIZE) {
			sprintf(LastErrorString , "ERROR:(GOTO)Invalid address %04X specified in address %04x.",
				pOp->op1, address);
			ret = -1;
		} else if (pMod->romType == C508 && pOp->op1 >= C508_ROMSIZE) {
			sprintf(LastErrorString , "WARNING:(GOTO)Invalid address %04X specified in address %04x.",
				pOp->op1, address);
			if (ret == 0) ret = 1;
		} else pMod->rom[pOp->op1].flags |= FLAG_JUMPED;
		break;

	case 0x0c:
		/*
		 *	MOVLW	k	1100 kkkk kkkk
		 */
		pOp->mnemonic = MN_MOVLW;
		pOp->op1 = lower;
		pOp->op2 = OPERAND_UNUSED;
		break;

	case 0x0d:
		/*
		 *	IORLW	k	1101 kkkk kkkk
		 */
		pOp->mnemonic = MN_IORLW;
		pOp->op1 = lower;
		pOp->op2 = OPERAND_UNUSED;
		break;

	case 0x0e:
		/*
		 *	ANDLW	k	1110 kkkk kkkk
		 */
		pOp->mnemonic = MN_ANDLW;
		pOp->op1 = lower;
		pOp->op2 = OPERAND_UNUSED;
		break;

	case 0x0f:
		/*
		 *	XORLW	k	1111 kkkk kkkk
		 */
		pOp->mnemonic = MN_XORLW;
		pOp->op1 = lower;
		pOp->op2 = OPERAND_UNUSED;
		break;

	default:
		sprintf(LastErrorString ,
			"ERROR: Invalid opcode %0x in address %04X.",
			opcode,
			address
			);
		pOp->mnemonic = MN_UNKNOWN;
		pOp->op1 = pOp->op2 = OPERAND_UNUSED;
		ret = -1;
		break;
	}

	return ret;
}

/*
 *	Disassemble the address and stores it into dest.
 *	This function is used only for test.
 */
int PicPrint(char *dest, PIC12C5XX *pMod, int address)
{
	OPCODE *pOp;

	if (address < 0 || address >= C509_ROMSIZE) {
		/* program error ??? */
		return 0;
	}
	pOp = &pMod->rom[address];
	if (!(pOp->flags & FLAG_USED)) return 0;

	sprintf(dest , "%04X: %03X\t%s\t" , address, pOp->code, Mnemonics[pOp->mnemonic]);
	if (pOp->op1 != OPERAND_UNUSED)
		sprintf(&dest[strlen(dest)], "%02Xh", pOp->op1);
	if (pOp->op2 != OPERAND_UNUSED)
		sprintf(&dest[strlen(dest)], ",%d", pOp->op2);
	strcat(dest , "\t;\n");

	return strlen(dest);
}