/*
 *	PIC 12C5xx DISASSEMBLER FOR MODCHIP Ver0.1
 *	COPYRIGHT(C) 1999 Hanimar
 *	This program can be freely distributed.
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "pic12c5x.h"

static void printSignature(PIC12C5XX *pMod, FILE *fpout);
static void printIncludes(PIC12C5XX *pMod, FILE *fpout);
static void printConfig(PIC12C5XX *pMod, FILE *fpout);
static void printCircuit(PIC12C5XX *pMod, FILE *fpout);
static void printVariables(PIC12C5XX *pMod, FILE *fpout);
static void printAsm(PIC12C5XX *pMod, int addr, FILE *fpout);
static void printFooter(PIC12C5XX *pMod, FILE *fpout);

static char *getRegisterName(char *dest, int no);
static char *getCharBit(char *dest , int no);

static void _type_f_d(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _type_f(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _type_f_b(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _type_k(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _type_noarg(PIC12C5XX *pMod, int addr, FILE *fpout);


static void _unknown(PIC12C5XX *pMod, int addr, FILE *fpout);

static void _addwf(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _andwf(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _clrf(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _clrw(PIC12C5XX *pMod, int addr, FILE *fpout);

static void _comf(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _decf(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _decfsz(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _incf(PIC12C5XX *pMod, int addr, FILE *fpout);

static void _incfsz(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _iorwf(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _movf(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _movwf(PIC12C5XX *pMod, int addr, FILE *fpout);

static void _nop(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _rlf(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _rrf(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _subwf(PIC12C5XX *pMod, int addr, FILE *fpout);

static void _swapf(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _xorwf(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _bcf(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _bsf(PIC12C5XX *pMod, int addr, FILE *fpout);

static void _btfsc(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _btfss(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _andlw(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _call(PIC12C5XX *pMod, int addr, FILE *fpout);

static void _clrwdt(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _goto(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _iorlw(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _movlw(PIC12C5XX *pMod, int addr, FILE *fpout);

static void _option(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _retlw(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _sleep(PIC12C5XX *pMod, int addr, FILE *fpout);
static void _tris(PIC12C5XX *pMod, int addr, FILE *fpout);

static void _xorlw(PIC12C5XX *pMod, int addr, FILE *fpout);


int PicPrintMore(PIC12C5XX *pMod, FILE *fpout)
{
	int i;
	int orgFlag = 1;
	int branchFlag = 0;

	printSignature(pMod, fpout);
	printCircuit(pMod, fpout);
	printIncludes(pMod, fpout);
	printConfig(pMod, fpout);
	printVariables(pMod, fpout);
	for (i = 0 ; i < C509_ROMSIZE ; i ++) {
		if (!(pMod->rom[i].flags & FLAG_USED)) {
			orgFlag = 1;
			continue;
		}
		if (orgFlag) {
			fprintf(fpout , "\n\n\torg\t0x%02X\n\n" , i);
			orgFlag = 0;
		}

		if (pMod->rom[i].flags & FLAG_CALLED) fprintf(fpout, "\nSUB_%03X\n" , i);
		if (pMod->rom[i].flags & FLAG_JUMPED) fprintf(fpout, "LAB_%03X\n" , i);

		printAsm(pMod, i, fpout);
		if (branchFlag) {
			fputc('\n' , fpout);
			branchFlag = 0;
		}
		if (pMod->rom[i].flags & FLAG_BRANCH) branchFlag = 1;
	}
	printFooter(pMod, fpout);
	
	return 0;
}

static void printSignature(PIC12C5XX *pMod, FILE *fpout)
{
	fprintf(fpout, ";Disassembled by %s,\n" , TARGET);
	fprintf(fpout , ";FILENAME:%s\n;\n;\n" , pMod->fname);
	fputs(";\n" , fpout);
	return;
}

static void printCircuit(PIC12C5XX *pMod, FILE *fpout)
{
	int i;
	int flags[7];

	for (i = 0 ; i < 7 ; i ++)
		flags[i] = (pMod->gpioFlag >> i) & 1;
	if (pMod->config != 0) {
		if (pMod->config & ~_MCLRE_OFF) flags[3] = 1;
		if (pMod->config & ~_INT_OSC) flags[5] = 1;
	}

	fputs(";                       _______  _______\n" , fpout);
	fputs(";                      |       \\/       |\n" , fpout);
	fprintf(fpout , ";                Vdd --+ 1 <<     >> 8  +-- Vss\n");
	fputs(";                      |                |\n" , fpout);

	fprintf(fpout , ";           GP5/OSC1 --+ 2 %s     %s 7  +-- GP0\n" ,
		flags[5] ? "<<" : "  ",
		flags[0] ? ">>" : "  ");
	fputs(";                      |                |\n" , fpout);

	fprintf(fpout , ";           GP4/OSC2 --+ 3 %s     %s 6  +-- GP1\n" ,
		flags[4] ? "<<" : "  ",
		flags[1] ? ">>" : "  ");
	fputs(";                      |                |\n" , fpout);

	fprintf(fpout , ";          GP3/!MCLR --+ 4 %s     %s 5  +-- GP2/T0CKI\n" ,
		flags[3] ? "<<" : "  ",
		flags[2] ? ">>" : "  ");
	fputs(";                      |                |\n" , fpout);
	fputs(";                      +----------------+\n" , fpout);

	return;

}


static void printIncludes(PIC12C5XX *pMod, FILE *fpout)
{
	fputs("P509    EQU     0       ;Default assembles for a PIC12C508\n\n" , fpout);
	fputs("\tIF      P509\n" , fpout);
	fputs("\tlist    p=12c509\n" , fpout);
	fputs("\tinclude \"p12c509.inc\"\n" , fpout);
	fputs("\tELSE\n" , fpout);
	fputs("\tlist    p=12c508\n" , fpout);
	fputs("\tinclude \"p12c508.inc\"\n" , fpout);
	fputs("\tENDIF\n" , fpout);
	fputs("\tradix   dec\n" , fpout);
	fputs(";\n" , fpout);
	return;
	pMod;	/* dummy */
}

static void printConfig(PIC12C5XX *pMod, FILE *fpout)
{
	int config = pMod->config;

	if (config == 0) return;	/* Couldn't detect configuration */

	fputs("\t__FUSES " , fpout);
	if (config & ~_MCLRE_OFF) fputs("_MCLRE_ON " , fpout);
	else fputs("_MCLRE_OFF " , fpout);

	if (config & ~_CP_ON) fputs("& _CP_OFF " , fpout);
	else fputs("& _CP_ON " , fpout);

	if (config & ~_WDT_OFF) fputs("& _WDT_ON " , fpout);
	else fputs("& _WDT_OFF " , fpout);

	if (!(config & ~_INT_OSC)) fputs("& _INT_OSC " , fpout);
	else if (!(config & ~_LP_OSC)) fputs("& _LP_OSC " , fpout);
	else if (!(config & ~_XT_OSC)) fputs("& _XT_OSC " , fpout);
	else fputs("& _RC_OSC " , fpout);

	fputs("\n\n" , fpout);
	return;
}

static void printVariables(PIC12C5XX *pMod, FILE *fpout)
{
	int min, max;
	int i;

	for (max = C509_RAMSIZE -1 ; max > 6 ; max --)
		if (pMod->ram[max] == RAM_USED) break;
	for (min = 7 ; min < max ; min ++)
		if (pMod->ram[min] == RAM_USED) break;

	fprintf(fpout, "\tcblock\t0x%02X\t;\n" , min);
	for (i = min ; i <= max ; i ++) {
		fprintf(fpout , "\t\tVAR_%02X\t;" , i);
		if (pMod->ram[i] == RAM_USED) fputc('\n' , fpout);
		else fputs("NOT USED\n" , fpout);
	}
	fputs("\tendc\n" , fpout);
}

static void printFooter(PIC12C5XX *pMod, FILE *fpout)
{
	fputs(";\n", fpout);
	fputs("; (Note: do NOT overwrite factory-programmed location 0x1FF/0x3FF !!)\n", fpout);
	fputs(";\n", fpout);
	fputs(";That's all, folks!\n", fpout);
	fputs(";\n", fpout);
	fputs("\tend\n" , fpout);
	return;
	pMod;	/* dummy */
}

static void printAsm(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	OPCODE *pOp;

	static void (*printFunctions[])(PIC12C5XX *pMod, int addr, FILE *fpout) = {
		_unknown,
		_addwf, _andwf, _clrf, _clrw,
		_comf, _decf, _decfsz, _incf,
		_incfsz, _iorwf, _movf, _movwf,
		_nop, _rlf, _rrf, _subwf,
		_swapf, _xorwf, _bcf, _bsf,
		_btfsc, _btfss, _andlw, _call,
		_clrwdt, _goto, _iorlw, _movlw,
		_option, _retlw, _sleep, _tris,
		_xorlw
	};

	pOp = &pMod->rom[addr];
	if (pOp->mnemonic > MN_LAST || pOp->mnemonic <= 0) {
		_unknown(pMod, addr, fpout);
		return;
	}
	(printFunctions[pOp->mnemonic])(pMod, addr, fpout);
	return;
}


static char *getRegisterName(char *dest, int no)
{
	static const char *RegNames[] = {
		"INDF", "TMR0", "PCL", "STATUS",
		"FSR", "OSCCAL", "GPIO",
		NULL
	};

	if (no >= 0 && no <= REG_MAX) sprintf(dest , "%s\t", RegNames[no]);
	else sprintf(dest, "VAR_%02X\t", no);
	
	return dest;
}

/*
 *	When no is larger than 255 then it returns "".
 */
static char *getCharBit(char *dest , int no)
{
	int i;
	char *p = dest;

	if (no > 255) dest[0] = '\0';
	else {
		*p ++ = 'b';
		*p ++ = '\'';
		for (i = 0 ; i < 8 ; i ++)
			*p ++ = ((no >> (7 - i)) & 1) ? '1':'0';
		*p ++ = '\'';
		*p ++ = '\0';
	}
	return dest;
}

static void _unknown(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	fprintf(fpout, "\tUNKNOWN CODE %03x!!!\n" , pMod->rom[addr].code);
	return;
}

static void _type_f_d(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	char buf[32];
	fprintf(fpout , "\t%s\t%s,%c\t;",
		Mnemonics[pMod->rom[addr].mnemonic],	/* for future release... */
		getRegisterName(buf, pMod->rom[addr].op1),
		pMod->rom[addr].op2 ? 'F':'W'
		);
}

static void _type_f(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	char buf[32];
	fprintf(fpout , "\t%s\t%s\t;",
		Mnemonics[pMod->rom[addr].mnemonic],	/* for future release... */
		getRegisterName(buf, pMod->rom[addr].op1)
		);
}

static void _type_f_b(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	char buf[32];
	fprintf(fpout , "\t%s\t%s,%d\t;",
		Mnemonics[pMod->rom[addr].mnemonic],	/* for future release... */
		getRegisterName(buf, pMod->rom[addr].op1),
		pMod->rom[addr].op2
		);
}

static void _type_k(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	char buf[32];
	fprintf(fpout , "\t%s\t%d\t\t;H'%02X' %s",
		Mnemonics[pMod->rom[addr].mnemonic],	/* for future release... */
		pMod->rom[addr].op1,
		pMod->rom[addr].op1,
		getCharBit(buf, pMod->rom[addr].op1)
		);
}

static void _type_noarg(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	fprintf(fpout , "\t%s\t\t\t;",
		Mnemonics[pMod->rom[addr].mnemonic]	/* for future release... */
		);
}


static void _addwf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _andwf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _clrf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _clrw(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_noarg(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _comf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _decf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _decfsz(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _incf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _incfsz(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _iorwf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _movf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _movwf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f(pMod, addr, fpout);
	fputc('\n' , fpout);
}

static void _nop(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_noarg(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _rlf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _rrf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _subwf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _swapf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _xorwf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_d(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _bcf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_b(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _bsf(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_b(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _btfsc(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_b(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _btfss(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f_b(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _andlw(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_k(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _call(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	fprintf(fpout , "\t%s\tSUB_%03X\t\t;",
		Mnemonics[pMod->rom[addr].mnemonic],	/* for future release... */
		pMod->rom[addr].op1
		);
	fprintf(fpout, "PC%s%d", (-(addr+1)+pMod->rom[addr].op1 >= 0) ? "+":"", 
		-(addr+1)+pMod->rom[addr].op1);
	fputc('\n' , fpout);
}
static void _clrwdt(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_noarg(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _goto(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	fprintf(fpout , "\t%s\tLAB_%03X\t\t;",
		Mnemonics[pMod->rom[addr].mnemonic],	/* for future release... */
		pMod->rom[addr].op1
		);
	fprintf(fpout, "PC%s%d", (-(addr+1)+pMod->rom[addr].op1 >= 0) ? "+":"", 
		-(addr+1)+pMod->rom[addr].op1);
	fputc('\n' , fpout);
}

static void _iorlw(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_k(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _movlw(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	if (addr + 1 < C509_ROMSIZE && (pMod->rom[addr + 1].flags & FLAG_USED) &&
		(	pMod->rom[addr + 1].mnemonic == MN_TRIS ||
			pMod->rom[addr + 1].mnemonic == MN_OPTION ||
			pMod->rom[addr + 1].mnemonic == MN_MOVWF && pMod->rom[addr + 1].op1 == REG_GPIO)
		)
	{	/* print argument as binary */
		char buf[32];
		fprintf(fpout , "\t%s\t%s\t;H'%02X' %d",
			Mnemonics[pMod->rom[addr].mnemonic],	/* for future release... */
			getCharBit(buf, pMod->rom[addr].op1),
			pMod->rom[addr].op1,
			pMod->rom[addr].op1
			);
	} else _type_k(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _option(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_noarg(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _retlw(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_k(pMod, addr, fpout);
	if (isprint(pMod->rom[addr].op1))
		fprintf(fpout , " '%c'\n" , pMod->rom[addr].op1);
	else fputc('\n' , fpout);
}
static void _sleep(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_noarg(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _tris(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_f(pMod, addr, fpout);
	fputc('\n' , fpout);
}
static void _xorlw(PIC12C5XX *pMod, int addr, FILE *fpout)
{
	_type_k(pMod, addr, fpout);
	fputc('\n' , fpout);
}

